@@ -5,12 +5,11 @@ import os
 import signal
 import sqlite3
 import tempfile
-from contextlib import contextmanager
-from inspect import isgeneratorfunction
-from itertools import chain, groupby, product, islice, takewhile, dropwhile
-from pathlib import Path, PurePath
 from collections import deque
-from collections.abc import Iterable
+from contextlib import contextmanager
+from itertools import (chain, dropwhile, groupby, islice, product,
+                       takewhile, tee)
+from pathlib import Path, PurePath
 
 
 class Conn:
@@ -30,29 +29,65 @@ class Conn:
         self._dbconn = None
 
     def __getitem__(self, tname):
-        return _InstructionContainer(self, tname)
+        return Dics((self, tname))
 
     def __setitem__(self, tname, val):
         with _connect(self._dbfile) as c:
             self._dbconn = c
             _delete(c, tname)
-            _insert(c, tname, _inst2iter(val))
+            _insert(c, tname, _dics2iter(val))
 
 
-class _InstructionContainer:
-    """Holds instructions to transform a table by chaining methods.
-
-    This is a private class. Instead of directly creating an instance of
-    this class, use Conn.__getitem__
+class Dics:
+    """
     """
 
-    def __init__(self, conn, tname):
-        self._conn = conn
-        self._insts = [{
-            'cmd': 'fetch',
-            'tname': tname,
-        }]
-        self._genfns = [_build_initgen(conn, tname, [])]
+    def __init__(self, src):
+        if isinstance(src, tuple):
+            conn, tname = src
+
+            self._conn = conn
+            self._elts = None
+            self._genfns = [_build_initgen(conn, tname, [])]
+
+            self._insts = [{
+                'cmd': 'fetch',
+                'tname': tname,
+            }]
+
+        else:
+            self._conn = None
+            self._elts = list(src)
+
+            def gen():
+                yield from self._elts
+
+            self._genfns = [gen]
+            self._insts = [{
+                'cmd': 'elts'
+            }]
+
+    def __str__(self):
+        return "\n".join(str(r) for r in self.iter())
+
+    def __getitem__(self, col):
+        return [r[col] for r in self.iter()]
+
+    def chain(self, other):
+        pgen = self._genfns[-1]
+
+        def gen():
+            for r in pgen():
+                yield r
+            for r in _dics2iter(other):
+                yield r
+
+        self._genfns.append(gen)
+        self._insts.append({
+            'cmd': 'chain',
+            'other': other
+        })
+        return self
 
     # each column may contain asc or desc
     # ex) 'col1 desc', 'col2 asc'
@@ -69,7 +104,7 @@ class _InstructionContainer:
 
         pgen = self._genfns[-1]
 
-        def gen():
+        def gen_from_sql():
             try:
                 tmpdbfd, tmpdb = tempfile.mkstemp()
                 with _connect(tmpdb) as c:
@@ -81,14 +116,33 @@ class _InstructionContainer:
                 if Path(tmpdb).is_file():
                     os.remove(tmpdb)
 
-        self._genfns.append(gen)
+        def gen_simp():
+            xs = list(pgen())
+            for col in reversed(cols):
+                rev = False
+                cs = col.split()
+                if len(cs) == 2:
+                    c, desc = cs
+                    if desc.lower() == 'desc':
+                        rev = True
+                else:
+                    c = cs[0]
+                xs.sort(key=lambda r: r[c], reverse=rev)
+            yield from xs
+
+        if self._insts[0]['cmd'] == 'fetch':
+            self._genfns.append(gen_from_sql)
+        else:
+            self._genfns.append(gen_simp)
+
+        self._genfns.append(gen_from_sql)
         self._insts.append({
             'cmd': 'order',
             'cols': cols
         })
         return self
 
-    def _map_raw(self, fn):
+    def apply(self, fn):
         pgen = self._genfns[-1]
 
         def gen_single():
@@ -111,7 +165,7 @@ class _InstructionContainer:
             self._genfns.append(gen_single)
 
         self._insts.append({
-            'cmd': '_map_raw',
+            'cmd': 'apply',
             'fn': fn
         })
         return self
@@ -133,7 +187,7 @@ class _InstructionContainer:
         pgen = self._genfns[-1]
 
         def gen():
-            yield from _step(fn, pgen(), _inst2iter(other))
+            yield from _step(fn, pgen(), _dics2iter(other))
 
         self._genfns.append(gen)
         self._insts.append({
@@ -143,84 +197,57 @@ class _InstructionContainer:
         })
         return self
 
-    def concat(self, other):
-        pgen = self._genfns[-1]
-
-        def gen():
-            for r in pgen():
-                yield r
-            for r in _inst2iter(other):
-                yield r
-
-        self._genfns.append(gen)
-        self._insts.append({
-            'cmd': 'concat',
-            'other': other
-        })
-        return self
-
     def map(self, fn):
         if self._insts[-1]['cmd'] == 'group':
             def _fn2gen(f):
-                # when f is a generator
-                def gen1(_k, rs):
-                    yield from fn(list(rs))
-
-                # when f is an ordinary function
-                def gen2(_k, rs):
-                    x = f(list(rs))
+                def gen(_k, rs):
+                    x = f(Dics(rs))
                     if isinstance(x, dict):
                         yield x
-                    elif x is not None:
-                        yield from x
-                return gen1 if isgeneratorfunction(f) else gen2
-            return self._map_raw(_fn2gen(fn))
+                    elif isinstance(x, Dics):
+                        yield from x.iter()
+                return gen
+            return self.apply(_fn2gen(fn))
 
         elif self._insts[-1]['cmd'] == 'windowed':
             def _fn2gen1(f):
-                # when f is an ordinary function
+
                 def gen(rs):
-                    x = f(rs)
+                    x = f(Dics(rs))
                     if isinstance(x, dict):
                         yield x
-                    elif x is not None:
-                        yield from x
-                return f if isgeneratorfunction(f) else gen
+                    elif isinstance(x, Dics):
+                        yield from x.iter()
+                return gen
 
             def _fn2gen2(f):
-                # when f is a generator
-                def gen1(rs):
+                def gen(rs):
                     rs = rs[:step] + [r.copy() for r in rs[step:]]
-                    yield from fn(rs)
-
-                # when f is an ordinary function
-                def gen2(rs):
-                    rs = rs[:step] + [r.copy() for r in rs[step:]]
-                    x = f(rs)
+                    x = f(Dics(rs))
                     if isinstance(x, dict):
                         yield x
-                    elif x is not None:
-                        yield from x
-                return gen1 if isgeneratorfunction(f) else gen2
+                    elif isinstance(x, Dics):
+                        yield from x.iter()
+                return gen
 
             step = self._insts[-1]['step']
             n = self._insts[-1]['n']
 
             if step >= n:
-                # no need to copy dicts
-                return self._map_raw(_fn2gen1(fn))
-            return self._map_raw(_fn2gen2(fn))
-        
+                # no need to copy
+                return self.apply(_fn2gen1(fn))
+            return self.apply(_fn2gen2(fn))
+
         # single row
-        def _fn2gen(f):
+        def _fn2gen_single(f):
             def gen(r):
                 x = f(r)
                 if isinstance(x, dict):
                     yield x
-                elif x is not None:
-                    yield from x
-            return f if isgeneratorfunction(f) else gen
-        return self._map_raw(_fn2gen(fn))
+                elif isinstance(x, Dics):
+                    yield from x.iter()
+            return gen
+        return self.apply(_fn2gen_single(fn))
 
     def by(self, *cols):
         # need to cut out 'desc', 'asc'
@@ -232,7 +259,7 @@ class _InstructionContainer:
             if pred(r):
                 yield r
 
-        return self.map(gen)
+        return self.apply(gen)
 
     def update(self, **kwargs):
         def updatefn(r):
@@ -240,10 +267,10 @@ class _InstructionContainer:
                 r.update({k: v(r)})
             yield r
 
-        return self.map(updatefn)
+        return self.apply(updatefn)
 
     # pretty expensive for what it actually does
-    # but this version does not depend on the order of rows. 
+    # but this version does not depend on the order of rows.
     def rename(self, **kwargs):
         kwargs_rev = {oldkey: newkey for newkey, oldkey in kwargs.items()}
 
@@ -251,8 +278,7 @@ class _InstructionContainer:
             yield {kwargs_rev.get(oldkey, oldkey): v
                    for oldkey, v in r.items()}
 
-        return self.map(renamefn)
-
+        return self.apply(renamefn)
 
     def fold(self, **kwargs):
         gcols = []
@@ -260,40 +286,31 @@ class _InstructionContainer:
             gcols = self._insts[-1]['cols']
 
         def foldfn1(rs):
+            ds = Dics(rs)
             r = {}
             for k, fn in kwargs.items():
-                val = fn(rs)
-                if isinstance(val, list):
-                    for i, v in enumerate(val):
-                        r[k + str(i)] = v
-                else:
-                    r[k] = val
+                r[k] = fn(ds)
             yield r
 
         def foldfn2(keys, rs):
-            rs = list(rs)
+            ds = Dics(rs)
             r = dict(zip(gcols, keys))
             for k, fn in kwargs.items():
-                val = fn(rs)
-                if not isinstance(val, str) and isinstance(val, Iterable):
-                    for i, v in enumerate(val):
-                        r[k + str(i)] = v
-                else:
-                    r[k] = val
+                r[k] = fn(ds)
             yield r
 
         if self._insts[-1]['cmd'] == 'windowed':
-            return self._map_raw(foldfn1)
+            return self.apply(foldfn1)
         elif self._insts[-1]['cmd'] == 'group':
-            return self._map_raw(foldfn2)
+            return self.apply(foldfn2)
         else:
-            raise ValueError("must be grouped or windowed before fold")
+            raise ValueError("fold must be preceded by grouping processes")
 
     def select(self, *cols):
         def selectfn(r):
             yield {col: r[col] for col in cols}
 
-        return self.map(selectfn)
+        return self.apply(selectfn)
 
     def deselect(self, *cols):
         cols = set(cols)
@@ -301,7 +318,7 @@ class _InstructionContainer:
         def deselectfn(r):
             yield {k: v for k, v in r.items() if k not in cols}
 
-        return self.map(deselectfn)
+        return self.apply(deselectfn)
 
     def inner_join(self, other):
         def innerjoin_fn(rs1, rs2, _left, _right):
@@ -369,7 +386,7 @@ class _InstructionContainer:
         def distinctfn(_k, rs):
             yield next(rs)
 
-        return self.by(*columns)._map_raw(distinctfn)
+        return self.by(*columns).apply(distinctfn)
 
     def windowed(self, n, step):
         pgen = self._genfns[-1]
@@ -431,6 +448,14 @@ class _InstructionContainer:
     def list(self):
         return list(self.iter())
 
+    # experimental
+    def tee(self, n):
+        return tuple(Dics(x) for x in tee(self.iter(), n))
+
+    # experimental
+    def new(self):
+        return Dics([r.copy() for r in self.iter()])
+
 
 def _insert_statement(name, d):
     """insert into foo values (:a, :b, :c, ...)
@@ -538,8 +563,9 @@ def _step(fn, krs1, krs2):
             yield from fn([], rs2, left, right)
 
 
-def _inst2iter(obj):
-    return obj.iter() if isinstance(obj, _InstructionContainer) else iter(obj)
+# InstructionSet (It) to iter
+def _dics2iter(obj):
+    return obj.iter() if isinstance(obj, Dics) else iter(obj)
 
 
 def _build_initgen(conn, tname, cols):
