@@ -5,47 +5,80 @@ from pathlib import Path
 
 from tablemap import Conn
 
-table1 = [
-    {'col1': 'a', 'col2': 4},
-    {'col1': 'a', 'col2': 5},
-    {'col1': 'b', 'col2': 1},
-    {'col1': 'c', 'col2': 3},
-    {'col1': 'c', 'col2': 4},
-    {'col1': 'c', 'col2': 7},
-    {'col1': 'd', 'col2': 2},
 
-]
-
-table2 = [
-    {'col1': 'd', 'col3': 3},
-    {'col1': 'b', 'col3': 1},
-    {'col1': 'e', 'col3': 3},
-    {'col1': 'e', 'col3': 4},
-    {'col1': 'd', 'col3': 2},
-]
-
-
-class TestMisc(unittest.TestCase):
+class TestCore(unittest.TestCase):
     def setUp(self):
         self.dbfile = 'test/sample.db'
+        self.dbfile1 = 'test/sample1.db'
+
+        self.t1 = [
+            {'col1': 'a', 'col2': 4},
+            {'col1': 'a', 'col2': 5},
+            {'col1': 'b', 'col2': 1},
+            {'col1': 'c', 'col2': 3},
+            {'col1': 'c', 'col2': 4},
+            {'col1': 'c', 'col2': 7},
+            {'col1': 'd', 'col2': 2},
+
+        ]
+
+        self.t2 = [
+            {'col1': 'd', 'col3': 3},
+            {'col1': 'b', 'col3': 1},
+            {'col1': 'e', 'col3': 3},
+            {'col1': 'e', 'col3': 4},
+            {'col1': 'd', 'col3': 2},
+        ]
 
     def tearDown(self):
         if Path(self.dbfile).is_file():
             os.remove(self.dbfile)
+        if Path(self.dbfile1).is_file():
+            os.remove(self.dbfile1)
+
+    def test_tee_n_new(self):
+        conn = Conn(self.dbfile)
+
+        def tempfn(ds):
+            ds1, ds2 = ds.tee(2)
+            return ds1.order('col2 desc').chain(ds2.order('col2'))
+
+        conn['t1'] = self.t1
+        conn['t1_1'] = conn['t1'].by('col1').map(tempfn)
+        self.assertEqual(conn['t1_1']['col2'],
+                         [5, 4, 4, 5, 1, 1, 7, 4, 3, 3, 4, 7, 2, 2])
+
+        def tempfn(ds):
+            ds1, ds2 = ds.tee(2)
+            ds1.update(col2=lambda r: r['col2'] + 1)
+            return ds1.order('col2 desc').chain(ds2.order('col2'))
+
+        conn['t1'] = self.t1
+        conn['t1_1'] = conn['t1'].by('col1').map(tempfn)
+        self.assertEqual(conn['t1_1']['col2'],
+                         [6, 5, 5, 6, 2, 2, 8, 5, 4, 4, 5, 8, 3, 3])
+
+        def tempfn(ds):
+            ds1, ds2 = ds.tee(2)
+            ds1.update(col2=lambda r: r['col2'] + 1)
+            return ds1.order('col2 desc').chain(ds2.new().order('col2'))
+
+        conn['t1'] = self.t1
+        conn['t1_1'] = conn['t1'].by('col1').map(tempfn)
+        self.assertEqual(conn['t1_1']['col2'],
+                         [6, 5, 4, 5, 2, 1, 8, 5, 4, 3, 4, 7, 3, 2])
 
     def test_index(self):
         conn = Conn(self.dbfile)
-        conn['t1'] = table1
+        conn['t1'] = self.t1
 
         def cnt():
             n = 0
 
-            def innerfn(rs):
+            def innerfn(ds):
                 nonlocal n
                 n += 1
-                for r in rs:
-                    r['n'] = n
-                    yield r
+                return ds.update(n=lambda r: n)
             return innerfn
 
         conn['t1_1'] = conn['t1'].by('col1').map(cnt())
@@ -58,15 +91,6 @@ class TestMisc(unittest.TestCase):
                                   {'col1': 'c', 'col2': 7, 'n': 3},
                                   {'col1': 'd', 'col2': 2, 'n': 4}])
 
-
-class TestFold(unittest.TestCase):
-    def setUp(self):
-        self.dbfile = 'test/sample.db'
-
-    def tearDown(self):
-        if Path(self.dbfile).is_file():
-            os.remove(self.dbfile)
-
     def test_fold1(self):
         conn = Conn(self.dbfile)
 
@@ -76,10 +100,11 @@ class TestFold(unittest.TestCase):
                 return [min(xs), max(xs)]
             return fn
 
-        conn['t1'] = table1
+        conn['t1'] = self.t1
         conn['t1_1'] = conn['t1'].by('col1').fold(
-            sum=lambda rs: sum(r['col2'] for r in rs),
-            mm=minmax('col2')
+            sum=lambda ds: sum(ds['col2']),
+            mm0=lambda ds: min(ds['col2']),
+            mm1=lambda ds: max(ds['col2'])
         )
 
         self.assertEqual(conn['t1_1'].list(),
@@ -89,8 +114,9 @@ class TestFold(unittest.TestCase):
                           {'col1': 'd', 'sum': 2, 'mm0': 2, 'mm1': 2}])
 
         conn['t1_1'] = conn['t1'].windowed(4, 3).fold(
-            sum=lambda rs: sum(r['col2'] for r in rs),
-            mm=minmax('col2')
+            sum=lambda ds: sum(ds['col2']),
+            mm0=lambda ds: min(ds['col2']),
+            mm1=lambda ds: max(ds['col2'])
         )
         self.assertEqual(conn['t1_1'].list(),
                          [{'sum': 13, 'mm0': 1, 'mm1': 5},
@@ -98,8 +124,9 @@ class TestFold(unittest.TestCase):
                           ])
 
         conn['t1_1'] = conn['t1'].windowed(4, 5).fold(
-            sum=lambda rs: sum(r['col2'] for r in rs),
-            mm=minmax('col2')
+            sum=lambda ds: sum(ds['col2']),
+            mm0=lambda ds: min(ds['col2']),
+            mm1=lambda ds: max(ds['col2'])
         ).rename(
             mm1='mm0',
             mm0='mm1'
@@ -109,24 +136,13 @@ class TestFold(unittest.TestCase):
                           {'sum': 9, 'mm1': 2, 'mm0': 7}
                           ])
 
-
-class TestGeneral(unittest.TestCase):
-    def setUp(self):
-        self.dbfile = 'test/sample.db'
-
-    def tearDown(self):
-        if Path(self.dbfile).is_file():
-            os.remove(self.dbfile)
-
     def test_by(self):
-        def sum_by_col1(rs):
-            r = {}
-            r['col1'] = rs[0]['col1']
-            r['col2'] = sum(r['col2'] for r in rs)
-            yield r
+        def sum_by_col1(ds):
+            tot = sum(ds['col2'])
+            return ds.update(col2=lambda r: tot).islice(1)
 
         conn = Conn(self.dbfile)
-        conn['t1'] = table1
+        conn['t1'] = self.t1
         conn['t1_sum'] = conn['t1'].by('col1')\
             .map(sum_by_col1)
 
@@ -141,52 +157,36 @@ class TestGeneral(unittest.TestCase):
 
         self.assertEqual(conn['t1_sum'].list(), conn['t1_sum2'].list())
 
-    def test_concat(self):
+    def test_chain(self):
         conn = Conn(self.dbfile)
-        conn['t1'] = table1
-        conn['t1_double'] = conn['t1'].concat(conn['t1'])
+        conn['t1'] = self.t1
+        conn['t1_double'] = conn['t1'].chain(conn['t1'])
         self.assertEqual(conn['t1'].list() + conn['t1'].list(),
                          conn['t1_double'].list())
 
-        conn['t1_double'] = conn['t1'].concat(table1)
+        conn['t1_double'] = conn['t1'].chain(self.t1)
         self.assertEqual(conn['t1'].list() + conn['t1'].list(),
                          conn['t1_double'].list())
 
     def test_df(self):
         import pandas as pd
         conn = Conn(self.dbfile)
-        conn['t1'] = table1
+        conn['t1'] = self.t1
         df = pd.DataFrame(conn['t1'].iter())
         conn['t1_copy'] = df.to_dict('records')
         self.assertEqual(conn['t1'].list(), conn['t1_copy'].list())
 
-
-class TestBy(unittest.TestCase):
-    def setUp(self):
-        self.dbfile = 'test/sample.db'
-        self.dbfile1 = 'test/sample1.db'
-
-    def tearDown(self):
-        if Path(self.dbfile).is_file():
-            os.remove(self.dbfile)
-        if Path(self.dbfile1).is_file():
-            os.remove(self.dbfile1)
-
     def test_by_after_map(self):
-        def sumit(rs):
-            tot = sum(r['col2']for r in rs)
-            for r in rs:
-                r['tot'] = tot
-                yield r
+        def sumit(ds):
+            tot = sum(ds['col2'])
+            return ds.update(tot=lambda r: tot)
 
-        def count(rs):
-            n = len(rs)
-            for r in rs:
-                r['cnt'] = n
-                yield r
+        def count(ds):
+            n = len(ds.list())
+            return ds.update(cnt=lambda r: n)
 
         conn = Conn(self.dbfile)
-        conn['t1'] = table1
+        conn['t1'] = self.t1
         conn['t1_sum'] = conn['t1'].by('col1')\
             .map(sumit).by('tot').map(count)
 
@@ -200,22 +200,20 @@ class TestBy(unittest.TestCase):
         self.assertEqual(conn['t1_sum'].list(), conn1['t1_sum2'].list())
 
     def test_multiple_by(self):
-        def foo(r):
-            r['col3'] = 1
-            return r
 
         conn = Conn(self.dbfile)
-        conn['t1'] = table1
-        conn['t1_1'] = conn['t1'].map(foo).by('col1', 'col3')\
-            .map(lambda rs: rs)
+        conn['t1'] = self.t1
+        conn['t1_1'] = conn['t1'].update(col3=lambda r: 1).by('col1', 'col3')\
+            .map(lambda ds: ds)
 
-        conn['t1_2'] = conn['t1'].map(foo).by('col1').map(lambda rs: rs)
+        conn['t1_2'] = conn['t1'].update(
+            col3=lambda r: 1).by('col1').map(lambda ds: ds)
         self.assertEqual(conn['t1_1'].list(), conn['t1_2'].list())
 
         conn['t1_1_double'] = conn['t1_1']\
-            .concat(conn['t1'].map(foo).by('col1', 'col3')
-                              .map(lambda rs: rs))
-        conn['t1_1_double2'] = conn['t1_1'].concat(conn['t1_1'])
+            .chain(conn['t1'].update(col3=lambda r: 1).by('col1', 'col3')
+                   .map(lambda rs: rs))
+        conn['t1_1_double2'] = conn['t1_1'].chain(conn['t1_1'])
 
         self.assertEqual(conn['t1_1_double'].list(),
                          conn['t1_1_double2'].list())
@@ -223,8 +221,8 @@ class TestBy(unittest.TestCase):
     def test_islice(self):
 
         conn = Conn(self.dbfile)
-        conn['t1'] = table1
-        conn['t2'] = table2
+        conn['t1'] = self.t1
+        conn['t2'] = self.t2
 
         self.assertEqual(len(conn['t1'].islice(3).list()), 3)
         self.assertEqual(len(conn['t1'].islice(3, None).list()), 4)
@@ -243,7 +241,7 @@ class TestBy(unittest.TestCase):
 
     def test_takewhile(self):
         conn = Conn(self.dbfile)
-        conn['t1'] = table1
+        conn['t1'] = self.t1
 
         self.assertEqual(len(conn['t1']
                              .takewhile(lambda r: r['col2'] > 1).list()), 2)
@@ -254,7 +252,7 @@ class TestBy(unittest.TestCase):
 
     def test_dropwhile(self):
         conn = Conn(self.dbfile)
-        conn['t1'] = table1
+        conn['t1'] = self.t1
 
         self.assertEqual(len(conn['t1']
                              .dropwhile(lambda r: r['col2'] > 1).list()), 5)
@@ -265,12 +263,10 @@ class TestBy(unittest.TestCase):
 
     def test_windowed(self):
         conn = Conn(self.dbfile)
-        conn['t1'] = table1
+        conn['t1'] = self.t1
 
-        def add1(rs):
-            for r in rs:
-                r['col2'] += 1
-                yield r
+        def add1(ds):
+            return ds.update(col2=lambda r: r['col2'] + 1)
 
         self.assertEqual(conn['t1'].windowed(4, 2).map(add1).list(),
                          [{'col1': 'a', 'col2': 5},
@@ -285,26 +281,17 @@ class TestBy(unittest.TestCase):
                           {'col1': 'c', 'col2': 8},
                           {'col1': 'd', 'col2': 3}])
 
-
-class TestDerivs(unittest.TestCase):
-    def setUp(self):
-        self.dbfile = 'test/sample.db'
-
-    def tearDown(self):
-        if Path(self.dbfile).is_file():
-            os.remove(self.dbfile)
-
     def test_filter(self):
         conn = Conn(self.dbfile)
 
-        conn['t1'] = table1
+        conn['t1'] = self.t1
         conn['t1_1'] = conn['t1'].filter(lambda r: r['col2'] > 2)
 
         self.assertEqual(len(conn['t1_1'].list()), 5)
 
     def test_update(self):
         conn = Conn(self.dbfile)
-        conn['t1'] = table1
+        conn['t1'] = self.t1
         conn['t1_1'] = conn['t1'].update(
             col2=lambda r: r['col2'] + 1,
             col3=lambda r: r['col1'] + str(r['col2'])
@@ -315,7 +302,11 @@ class TestDerivs(unittest.TestCase):
 
     def test_select(self):
         conn = Conn(self.dbfile)
-        conn['t1'] = table1
+        conn['t1'] = self.t1
+
+        # for r in conn['t1'].iter():
+        #     print(r)
+
         conn['t1_1'] = conn['t1'].update(
             col2=lambda r: r['col2'] + 1,
             col3=lambda r: r['col1'] + str(r['col2'])
@@ -330,8 +321,8 @@ class TestDerivs(unittest.TestCase):
 
     def test_join(self):
         conn = Conn(self.dbfile)
-        conn['t1'] = table1
-        conn['t2'] = table2
+        conn['t1'] = self.t1
+        conn['t2'] = self.t2
 
         conn['t1_1'] = conn['t1'].by('col1').inner_join(conn['t2'].by('col1'))
 
@@ -362,10 +353,10 @@ class TestDerivs(unittest.TestCase):
 
     def test_join2(self):
         conn = Conn(self.dbfile)
-        conn['t1'] = table1
-        conn['t2'] = table2
+        conn['t1'] = self.t1
+        conn['t2'] = self.t2
 
-        conn['t1_1'] = conn['t1'].concat([{'col1': 'd', 'col2': 9}, {
+        conn['t1_1'] = conn['t1'].chain([{'col1': 'd', 'col2': 9}, {
             'col1': 'd', 'col2': 19}]).by('col1')\
             .full_join(conn['t2'].by('col1'))
 
@@ -387,24 +378,15 @@ class TestDerivs(unittest.TestCase):
 
     def test_distinct(self):
         conn = Conn(self.dbfile)
-        conn['t1'] = table1
+        conn['t1'] = self.t1
         conn['t1_1'] = conn['t1'].distinct('col1')
         self.assertEqual(len(conn['t1_1'].list()), 4)
         conn['t1_2'] = conn['t1'].distinct('col1', 'col2')
         self.assertEqual(len(conn['t1_2'].list()), 7)
 
-
-class TestOrder(unittest.TestCase):
-    def setUp(self):
-        self.dbfile = 'test/sample.db'
-
-    def tearDown(self):
-        if Path(self.dbfile).is_file():
-            os.remove(self.dbfile)
-
     def test_order(self):
         conn = Conn(self.dbfile)
-        conn['t1'] = table1
+        conn['t1'] = self.t1
 
         [g1, g2, g3, g4] = [list(xs) for _, xs in conn['t1'].order(
             'col1', 'col2 desc').group('col1').iter()]
@@ -416,22 +398,14 @@ class TestOrder(unittest.TestCase):
                          'col1': 'c', 'col2': 4}, {'col1': 'c', 'col2': 3}])
         self.assertEqual(g4, [{'col1': 'd', 'col2': 2}])
 
-
-class TestReadMe(unittest.TestCase):
-
-    def tearDown(self):
-        if Path('test/sample.db').is_file():
-            os.remove('test/sample.db')
-
-    def test_all_of_em(self):
-
+    def test_readme(self):
         t1 = [
             {'col1': 'a', 'col2': 4},
             {'col1': 'a', 'col2': 5},
             {'col1': 'b', 'col2': 1},
         ]
 
-        conn = Conn('test/sample.db')
+        conn = Conn(self.dbfile)
         conn['t1'] = t1
 
         for r1, r2 in zip(conn['t1'].list(), conn['t1'].iter()):
@@ -444,36 +418,29 @@ class TestReadMe(unittest.TestCase):
 
         self.assertEqual(conn['t1'].list(), conn['t1_copy'].list())
 
-        def add1(r):
-            if r['col2'] > 2:
-                r['col2'] += 1
-                return r
-
-        conn['t1_1'] = conn['t1'].map(add1)
+        conn['t1_1'] = conn['t1']\
+            .filter(lambda r: r['col2'] > 2)\
+            .update(col2=lambda r: r['col2'] + 1)
 
         self.assertEqual(conn['t1_1'].list()[0]['col2'], 5)
 
-        def sumup(rs):
-            r0 = rs[0]
-            r0['col2_sum'] = sum(r['col2'] for r in rs)
-            del r0['col2']
-            return r0
-
         conn['t1_col2sum_groupby_col1'] = conn['t1'].by('col1')\
-            .map(sumup)
+            .fold(col2_sum=lambda ds: sum(ds['col2']))\
+            .deselect('col2')
 
         conn['t1_col2sum_groupby_col1_1'] = conn['t1']\
             .order('col1', 'col2 desc').group('col1')\
-            .map(sumup)
+            .fold(col2_sum=lambda ds: sum(ds['col2']))\
+            .deselect('col2')
 
         self.assertEqual(conn['t1_col2sum_groupby_col1'].list(),
                          conn['t1_col2sum_groupby_col1_1'].list())
 
-        conn['t1_double'] = conn['t1'].concat(conn['t1'])
+        conn['t1_double'] = conn['t1'].chain(conn['t1'])
         self.assertEqual(len(conn['t1_double'].list()),
                          2 * len(conn['t1'].list()))
 
-        conn['t1_double'] = conn['t1'].concat(t1)
+        conn['t1_double'] = conn['t1'].chain(t1)
 
         self.assertEqual(len(conn['t1_double'].list()),
                          2 * len(conn['t1'].list()))
